#!/usr/bin/env python
"""
The imagecalc module contains functions for performing a number of
calculating on images.
"""

import math
import os
from typing import List, Union

import numpy
from osgeo import gdal

import rsgislib

# import the C++ extension into this level
from ._imagecalc import *

gdal.UseExceptions()

# define our own classes
class BandDefn(object):
    """
    Create a list of these objects to pass to the band_math function
    as the 'bands' parameter.
    """

    def __init__(self, band_name=None, input_img=None, img_band=None):
        self.band_name = band_name
        self.input_img = input_img
        self.img_band = img_band


class StatsSummary:
    """
    This is passed to the imagePixelColumnSummary function"""

    def __init__(
        self,
        min=0.0,
        max=0.0,
        sum=0.0,
        median=0.0,
        stdev=0.0,
        mean=0.0,
        mode=0.0,
        calc_min=False,
        calc_max=False,
        calc_sum=False,
        calc_mean=False,
        calc_stdev=False,
        calc_median=False,
        calc_mode=False,
    ):
        self.min = min
        self.max = max
        self.sum = sum
        self.mean = mean
        self.stdev = stdev
        self.median = median
        self.mode = mode
        self.calc_min = calc_min
        self.calc_max = calc_max
        self.calc_sum = calc_sum
        self.calc_mean = calc_mean
        self.calc_stdev = calc_stdev
        self.calc_median = calc_median
        self.calc_mode = calc_mode


class ImageBandRescale:
    """
    Data structure for rescaling information for rescale_img_pxl_vals function.
    :param band: specified image band (band numbering starts at 1).
    :param in_min: the input image band minimum value for rescaling.
    :param in_max: the input image band maximum value for rescaling.
    :param no_data_val: no data value for the input image band.
    :param out_min: the output image band minimum value for rescaling.
    :param out_max: the output image band maximum value for rescaling.
    :param out_no_data: no data value for the output image band.
    """

    def __init__(
        self,
        band=0,
        in_min=0.0,
        in_max=0.0,
        no_data_val=0,
        out_min=0.0,
        out_max=0.0,
        out_no_data=0.0,
    ):
        """
        :param band: specified image band (band numbering starts at 1).
        :param in_min: the input image band minimum value for rescaling.
        :param in_max: the input image band maximum value for rescaling.
        :param no_data_val: no data value for the input image band.
        :param out_min: the output image band minimum value for rescaling.
        :param out_max: the output image band maximum value for rescaling.
        :param out_no_data: no data value for the output image band.

        """
        self.band = band
        self.in_min = in_min
        self.in_max = in_max
        self.no_data_val = no_data_val
        self.out_min = out_min
        self.out_max = out_max
        self.out_no_data = out_no_data

    def __str__(self):
        str_val = "Band {} [{}, {}, {}] [{}, {}, {}]".format(
            self.band,
            self.in_min,
            self.in_max,
            self.no_data_val,
            self.out_min,
            self.out_max,
            self.out_no_data,
        )
        return str_val

    def __repr__(self):
        str_val = "Band {} [{}, {}, {}] [{}, {}, {}]".format(
            self.band,
            self.in_min,
            self.in_max,
            self.no_data_val,
            self.out_min,
            self.out_max,
            self.out_no_data,
        )
        return str_val


def calc_dist_to_img_vals(
    input_img: str,
    output_img: str,
    pxl_vals: List[int],
    img_band: int = 1,
    gdalformat: str = "KEA",
    max_dist: float = None,
    no_data_val: float = None,
    out_no_data_val: float = None,
    unit_geo: bool = True,
):
    """
    A function to calculate the distance to the nearest pixel value with one
    of the specified values.

    :param input_img: is a string specifying the input image file.
    :param output_img: is a string specfiying the output image file.
    :param pxl_vals: is a number of list of numbers specifying the features to which
                     the distance from should be calculated.
    :param img_band: is an integer specifying the image band of the input image to
                     be used (Default = 1).
    :param gdalformat: is a string specifying the output image format (Default = KEA)
    :param max_dist: is a number specifying the maximum distance to be calculated, if
                     None not max value is used (Default = None).
    :param no_data_val: is the no data value in the input image for which distance
                        should not be calculated for (Default = None; None = no
                        specified no data value).
    :param out_no_data_val: is output image no data value. If None then set as the
                            max_dist value.
    :param unit_geo: is a boolean specifying the output distance units.
                     True = Geographic units (e.g., metres),
                     False is in Pixels (Default = True).

    .. code:: python

        import rsgislib.imagecalc
        cloudsImg = 'LS5TM_20110701_lat52lon421_r24p204_clouds.kea'
        dist2Clouds = 'LS5TM_20110701_lat52lon421_r24p204_distclouds.kea'
        # Pixel value 1 == Clouds
        # Pixel value 2 == Cloud Shadows
        rsgislib.imagecalc.calc_dist_to_img_vals(cloudsImg, dist2Clouds, pxl_vals=[1,2], out_no_data_val=-9999)

    """
    # Check gdal is available
    import rsgislib.imageutils

    have_list_vals = False
    if isinstance(pxl_vals, list):
        have_list_vals = True

    proxOptions = []

    if max_dist is not None:
        proxOptions.append("MAXDIST=" + str(max_dist))
    if no_data_val is not None:
        proxOptions.append("NODATA=" + str(no_data_val))
    if unit_geo:
        proxOptions.append("DISTUNITS=GEO")
    else:
        proxOptions.append("DISTUNITS=PIXEL")

    if have_list_vals:
        strVals = ""
        first = True
        for val in pxl_vals:
            if first:
                strVals = str(val)
                first = False
            else:
                strVals = strVals + "," + str(val)
        proxOptions.append("VALUES=" + strVals)
    else:
        proxOptions.append("VALUES=" + str(pxl_vals))

    try:
        import tqdm

        pbar = tqdm.tqdm(total=100)
        callback = lambda *args, **kw: pbar.update()
    except:
        callback = gdal.TermProgress

    if out_no_data_val is None:
        out_no_data_val = max_dist
    if out_no_data_val is None:
        raise ValueError("Either out_no_data_val or max_dist must be supplied.")

    valsImgDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    valsImgBand = valsImgDS.GetRasterBand(img_band)
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, out_no_data_val, gdalformat, rsgislib.TYPE_32FLOAT
    )
    distImgDS = gdal.Open(output_img, gdal.GA_Update)
    distImgBand = distImgDS.GetRasterBand(1)
    gdal.ComputeProximity(valsImgBand, distImgBand, proxOptions, callback=callback)
    distImgBand = None
    distImgDS = None
    valsImgBand = None
    valsImgDS = None


def _computeProximityArrArgsFunc(arg_vals):
    """
    This function is used internally within calc_dist_to_classes
    for the multiprocessing Pool
    """
    import rsgislib.imageutils

    # 0: tileFile
    # 1: distTileFile
    # 2: proxOptions
    # 3: no_data_val,
    # 4: format (e.g., KEA)
    # 5: img_band
    try:
        import tqdm

        pbar = tqdm.tqdm(total=100)
        callback = lambda *args, **kw: pbar.update()
    except:
        callback = gdal.TermProgress
    classImgDS = gdal.Open(arg_vals[0], gdal.GA_ReadOnly)
    classImgBand = classImgDS.GetRasterBand(arg_vals[5])
    rsgislib.imageutils.create_copy_img(
        arg_vals[0], arg_vals[1], 1, arg_vals[3], arg_vals[4], rsgislib.TYPE_32FLOAT
    )
    distImgDS = gdal.Open(arg_vals[1], gdal.GA_Update)
    distImgBand = distImgDS.GetRasterBand(arg_vals[5])
    gdal.ComputeProximity(classImgBand, distImgBand, arg_vals[2], callback=callback)
    distImgBand = None
    distImgDS = None
    classImgBand = None
    classImgDS = None


def calc_dist_to_img_vals_tiled(
    input_img,
    output_img,
    pxl_vals,
    img_band=1,
    max_dist=1000,
    no_data_val=1000,
    out_no_data_val=None,
    gdalformat="KEA",
    unit_geo=True,
    tmp_dir="tmp",
    tile_size=2000,
    n_cores=-1,
):
    """
    A function to calculate the distance to the nearest pixel value with
    one of the specified values.

    :param input_img: is a string specifying the input image file.
    :param output_img: is a string specfiying the output image file.
    :param pxl_vals: is a number of list of numbers specifying the features to
                     which the distance from should be calculated.
    :param img_band: is an integer specifying the image band of the input image
                     to be used (Default = 1).
    :param gdalformat: is a string specifying the output image format (Default = KEA)
    :param max_dist: is a number specifying the maximum distance to be calculated,
                     if None not max value is used (Default = None).
    :param no_data_val: is the no data value in the input image for which distance
                        should not be calculated for (Default = None;
                        None = no specified no data value).
    :param out_no_data_val: is output image no data value. If None then set as the
                            max_dist value.
    :param unit_geo: is a boolean specifying the output distance units.
                     True = Geographic units (e.g., metres),
                     False is in Pixels (Default = True).
    :param tmp_dir: is a directory to be used for storing the image tiles and other
                    temporary files - if not directory does not exist it will be
                    created and deleted on completion (Default: tmp).
    :param tile_size: is an int specifying in pixels the size of the image tiles
                      used for processing (Default: 2000)
    :param n_cores: is the number of processing cores which are available to be used
                    for this processing. If -1 all available cores will be
                    used. (Default: -1)

    .. code:: python

        import rsgislib.imagecalc
        cloudsImg = 'LS5TM_20110701_lat52lon421_r24p204_clouds.kea'
        dist2Clouds = 'LS5TM_20110701_lat52lon421_r24p204_distclouds.kea'
        # Pixel value 1 == Clouds
        # Pixel value 2 == Cloud Shadows
        rsgislib.imagecalc.calc_dist_to_img_vals_tiled(cloudsImg, dist2Clouds, pxl_vals=[1,2])

    """

    # Check gdal is available
    import glob
    import shutil
    from multiprocessing import Pool
    import rsgislib.imageutils
    import rsgislib.tools.utils
    import rsgislib.tools.filetools

    have_list_vals = False
    if isinstance(pxl_vals, list):
        have_list_vals = True

    tmpPresent = True
    if not os.path.exists(tmp_dir):
        print(
            "WARNING: '{}' directory does not exist so creating it...".format(tmp_dir)
        )
        os.makedirs(tmp_dir)
        tmpPresent = False

    if n_cores <= 0:
        n_cores = rsgislib.tools.utils.num_process_cores()

    if out_no_data_val is None:
        out_no_data_val = max_dist

    uid = rsgislib.tools.utils.uid_generator()

    xRes, yRes = rsgislib.imageutils.get_img_res(input_img)
    if unit_geo:
        xMaxDistPxl = math.ceil(max_dist / xRes) + 10
        yMaxDistPxl = math.ceil(max_dist / yRes) + 10
    else:
        xMaxDistPxl = max_dist + 10
        yMaxDistPxl = max_dist + 10

    tileOverlap = xMaxDistPxl
    if yMaxDistPxl > xMaxDistPxl:
        tileOverlap = yMaxDistPxl

    imgTilesDIR = os.path.join(tmp_dir, "ImgTiles_" + uid)
    imgTilesDIRPresent = True
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
        imgTilesDIRPresent = False

    imgTileBase = os.path.join(imgTilesDIR, "ImgTile")
    rsgislib.imageutils.create_tiles(
        input_img,
        imgTileBase,
        tile_size,
        tile_size,
        tileOverlap,
        0,
        "KEA",
        rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
        "kea",
    )
    imgTileFiles = glob.glob(imgTileBase + "*.kea")

    distTilesDIR = os.path.join(tmp_dir, "DistTiles_" + uid)
    distTilesDIRPresent = True
    if not os.path.exists(distTilesDIR):
        os.makedirs(distTilesDIR)
        distTilesDIRPresent = False

    proxOptions = []

    if max_dist is not None:
        proxOptions.append("MAXDIST=" + str(max_dist))
    if no_data_val is not None:
        proxOptions.append("NODATA=" + str(no_data_val))
    if unit_geo:
        proxOptions.append("DISTUNITS=GEO")
    else:
        proxOptions.append("DISTUNITS=PIXEL")

    if have_list_vals:
        strVals = ""
        first = True
        for val in pxl_vals:
            if first:
                strVals = str(val)
                first = False
            else:
                strVals = strVals + "," + str(val)
        proxOptions.append("VALUES=" + strVals)
    else:
        proxOptions.append("VALUES=" + str(pxl_vals))

    distTiles = []
    distTileArgs = []
    for tileFile in imgTileFiles:
        baseTileName = os.path.basename(tileFile)
        distTileFile = os.path.join(distTilesDIR, baseTileName)
        tileArgs = [
            tileFile,
            distTileFile,
            proxOptions,
            out_no_data_val,
            "KEA",
            img_band,
        ]
        distTiles.append(distTileFile)
        distTileArgs.append(tileArgs)

    with Pool(n_cores) as p:
        p.map(_computeProximityArrArgsFunc, distTileArgs)

    # Mosaic Tiles
    rsgislib.imageutils.create_img_mosaic(
        distTiles, output_img, 0, 0, 1, 1, gdalformat, rsgislib.TYPE_32FLOAT
    )
    rsgislib.imageutils.pop_img_stats(
        output_img, use_no_data=True, no_data_val=out_no_data_val, calc_pyramids=True
    )

    for imgFile in distTiles:
        rsgislib.tools.filetools.delete_file_with_basename(imgFile)

    if not imgTilesDIRPresent:
        shutil.rmtree(imgTilesDIR, ignore_errors=True)
    else:
        for tileFile in imgTileFiles:
            rsgislib.tools.filetools.delete_file_with_basename(tileFile)

    if not distTilesDIRPresent:
        shutil.rmtree(distTilesDIR, ignore_errors=True)

    if not tmpPresent:
        shutil.rmtree(tmp_dir, ignore_errors=True)


def buffer_img_pxl_vals(
    input_img: str,
    output_img: str,
    pxl_vals: List[int],
    buf_thres: float,
    tmp_dir: str,
    gdalformat: str = "KEA",
    img_band: int = 1,
    unit_geo: bool = True,
):
    """
    A function which uses the calc_dist_to_img_vals function and a threshold to
    buffer image pixel value(s) to create a binary mask

    :param input_img: The input image
    :param output_img: the output image
    :param pxl_vals: a list of pixel values defining the region to be buffered
    :param buf_thres: the threshold below which the buffered region will be defined.
    :param tmp_dir: a tmp directory for intermediate outputs
    :param gdalformat: output image format
    :param img_band: the input image bands
    :param unit_geo: is a boolean specifying the buf_thres distance units.
                     True = Geographic units (e.g., metres),
                     False is in Pixels (Default = True).

    """
    import rsgislib.tools.filetools

    if not os.path.exists(tmp_dir):
        raise rsgislib.RSGISPyException("The tmp_dir does not exist.")

    basename = rsgislib.tools.filetools.get_file_basename(input_img)

    dist_img = os.path.join(tmp_dir, f"{basename}_dist.kea")

    calc_dist_to_img_vals(
        input_img,
        dist_img,
        pxl_vals,
        img_band,
        "KEA",
        max_dist=(buf_thres * 1.1),
        no_data_val=None,
        out_no_data_val=-999,
        unit_geo=unit_geo,
    )

    image_math(
        dist_img,
        output_img,
        "(b1>=0)&&(b1<{})?1:0".format(buf_thres),
        gdalformat,
        rsgislib.TYPE_32FLOAT,
    )

    rsgislib.tools.filetools.delete_file_silent(dist_img)


def count_pxls_of_val(input_img: str, vals: List[int], img_band: int = None):
    """
    Function which counts the number of pixels of a set of values returning a
    list in the same order as the list of values provided.

    :param input_img: the input image
    :param vals: is a list of pixel values to be counted
    :param img_band: specify the image band for which the analysis is to be undertaken.
                     If None (default) then all bands will be used.
    :return: list of pixel counts in same order as the vals input list

    """
    from rios.imagereader import ImageReader
    import rsgislib.imageutils

    if vals is None:
        raise rsgislib.RSGISPyException("Input vals list must not be None.")
    if len(vals) == 0:
        raise rsgislib.RSGISPyException(
            "At least 1 value should be provided within the vals input variable."
        )
    n_bands = rsgislib.imageutils.get_img_band_count(input_img)
    if (img_band is not None) and ((img_band < 1) or (img_band > n_bands)):
        raise rsgislib.RSGISPyException(
            "The specified input image band is not within the input image."
        )
    if img_band is not None:
        img_band_idx = img_band - 1

    numVals = len(vals)
    outVals = numpy.zeros(numVals, dtype=numpy.int64)

    reader = ImageReader(input_img)
    for (info, block) in reader:
        for idx in range(numVals):
            if img_band is None:
                outVals[idx] = outVals[idx] + (block == vals[idx]).sum()
            else:
                outVals[idx] = (
                    outVals[idx]
                    + (
                        block[
                            img_band_idx,
                        ]
                        == vals[idx]
                    ).sum()
                )

    return outVals


def get_unique_values(input_img, img_band=1):
    """
    Find the unique image values within an image band.
    Note, the whole image band gets read into memory.

    :param input_img: input image file path
    :param img_band: image band to be processed (starts at 1)

    :return: array of unique values.

    """
    imgDS = gdal.Open(input_img)
    if imgDS is None:
        raise rsgislib.RSGISPyException("Could not open output image")
    imgBand = imgDS.GetRasterBand(img_band)
    if imgBand is None:
        raise rsgislib.RSGISPyException(
            "Could not open output image band ({})".format(img_band)
        )
    valsArr = imgBand.ReadAsArray()
    imgDS = None

    uniq_vals = numpy.unique(valsArr)
    return uniq_vals


def get_pca_eigen_vector(
    input_img, pxl_n_sample, no_data_val=None, out_matrix_file=None
):
    """
    A function which takes a sample from an input image and uses it to
    generate eigenvector for a PCA. Note. this can be used as input
    to rsgislib.imagecalc.pca

    :param input_img: the image from which the random sample will be taken.
    :param pxl_n_sample: the sample to be taken (e.g., a value of 100 will
                         sample every 100th pixel)
    :param no_data_val: provide a no data value which is to be ignored during
                        processing. If None then ignored (Default: None)
    :param out_matrix_file: path and name for the output rsgislib matrix file.
                            If None file is not created (Default: None)

    :returns: 1. array with the eigenvector, 2. array with the ratio of the
              explained variance
    """
    from sklearn.decomposition import PCA
    import rsgislib.imageutils

    # Read input data from image file.
    X = rsgislib.imageutils.extract_img_pxl_sample(input_img, pxl_n_sample, no_data_val)

    print(str(X.shape[0]) + " values were extracted from the input image.")

    pca = PCA()
    pca.fit(X)

    if out_matrix_file is not None:
        f = open(out_matrix_file, "w")
        f.write("m=" + str(pca.components_.shape[0]) + "\n")
        f.write("n=" + str(pca.components_.shape[1]) + "\n")
        first = True
        for val in pca.components_.flatten():
            if first:
                f.write(str(val))
                first = False
            else:
                f.write("," + str(val))
        f.write("\n\n")
        f.flush()
        f.close()

    pcaComp = 1
    print("Prop. of variance explained:")
    for val in pca.explained_variance_ratio_:
        print("\t PCA Component " + str(pcaComp) + " = " + str(round(val, 4)))
        pcaComp = pcaComp + 1

    return pca.components_, pca.explained_variance_ratio_


def perform_image_pca(
    input_img,
    output_img,
    out_eigen_vec_file,
    n_comps=None,
    pxl_n_sample=100,
    gdalformat="KEA",
    datatype=rsgislib.TYPE_32FLOAT,
    no_data_val=None,
    calc_stats=True,
):
    """
    A function which performs a PCA on the input image.

    :param input_img: the image from which the random sample will be taken.
    :param output_img: the output image transformed using the calculated PCA
    :param out_eigen_vec_file: path and name for the output rsgislib matrix file
                               containing the eigenvector for the PCA.
    :param n_comps: the number of PCA compoents outputted. A value of None is
                    all components (Default: None)
    :param pxl_n_sample: the sample to be taken (e.g., a value of 100 will sample
                         every 100th pixel) (Default: 100)
    :param gdalformat: the output gdal supported file format (Default KEA)
    :param datatype: the output data type of the input image
                     (Default: rsgislib.TYPE_32FLOAT)
    :param no_data_val: provide a no data value which is to be ignored during
                        processing. If None then ignored (Default: None)
    :param calc_stats: Boolean specifying whether pyramids and statistics should be
                       calculated for the output image. (Default: True)
    :returns: an array with the ratio of the explained variance per band.

    """
    import rsgislib.imageutils

    eigenVec, varExplain = get_pca_eigen_vector(
        input_img, pxl_n_sample, no_data_val, out_eigen_vec_file
    )
    outNComp = varExplain.shape[0]
    if n_comps is not None:
        if n_comps > varExplain.shape[0]:
            raise rsgislib.RSGISPyException(
                "You cannot output more components than "
                "the number of input image bands."
            )
        outNComp = n_comps

    pca(input_img, out_eigen_vec_file, output_img, outNComp, gdalformat, datatype)
    if calc_stats:
        use_no_data = False
        lcl_no_data_val = 0
        if no_data_val is not None:
            use_no_data = True
            lcl_no_data_val = no_data_val
        rsgislib.imageutils.pop_img_stats(
            output_img, use_no_data, lcl_no_data_val, True
        )

    return varExplain


def perform_image_mnf(
    input_img,
    output_img,
    n_comps=None,
    pxl_n_sample=100,
    in_img_no_data=None,
    tmp_dir="tmp",
    gdalformat="KEA",
    datatype=rsgislib.TYPE_32FLOAT,
    calc_stats=True,
):
    """
    A function which takes a sample from an input image and uses it to
    generate eigenvector for a MNF. Note. this can be used as input to
    rsgislib.imagecalc.pca

    :param input_img: the image to which the MNF will be applied
    :param output_img: the output image file with the MNF result
    :param n_comps: the number of components to be outputted
    :param pxl_n_sample: the sample to be taken (e.g., a value of 100 will sample
                         every 100th pixel) for the PCA
    :param in_img_no_data: provide a no data value which is to be ignored during
                           processing. If None then try to read from input image.
    :param tmp_dir: a directory where temporary output files will be stored. If it
                    doesn't exist it will be created.
    :param gdalformat: output image file format
    :param datatype: data type for the output image. Note, it is common to have
                     negative values.
    :param calc_stats: whether image statistics and pyramids could be calculated.
    :returns: array with the ratio of the explained variance

    """
    from sklearn.decomposition import PCA
    import rsgislib.tools.filetools
    import rsgislib.imageutils
    import shutil

    created_tmp_dir = False
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)
        created_tmp_dir = True

    img_basename = rsgislib.tools.filetools.get_file_basename(input_img)

    if in_img_no_data is None:
        in_img_no_data = rsgislib.imageutils.get_img_no_data_value(input_img)
        if in_img_no_data is None:
            raise rsgislib.RSGISPyException(
                "A no data value for the input image must be provided."
            )

    valid_msk_img = os.path.join(tmp_dir, "{}_vld_msk.kea".format(img_basename))
    rsgislib.imageutils.gen_valid_mask(input_img, valid_msk_img, "KEA", in_img_no_data)

    whiten_img = os.path.join(tmp_dir, "{}_whiten.kea".format(img_basename))
    rsgislib.imageutils.whiten_image(input_img, valid_msk_img, 1, whiten_img, "KEA")

    # Read input data from image file.
    X = rsgislib.imageutils.extract_img_pxl_sample(
        whiten_img, pxl_n_sample, in_img_no_data
    )
    print("{} values were extracted from the input image.".format(X.shape[0]))

    pca_obj = PCA()
    pca_obj.fit(X)

    eigenVecFile = os.path.join(tmp_dir, "{}_eigen_vec.txt".format(img_basename))
    f = open(eigenVecFile, "w")
    f.write("m=" + str(pca_obj.components_.shape[0]) + "\n")
    f.write("n=" + str(pca_obj.components_.shape[1]) + "\n")
    first = True
    for val in pca_obj.components_.flatten():
        if first:
            f.write(str(val))
            first = False
        else:
            f.write("," + str(val))
    f.write("\n\n")
    f.flush()
    f.close()

    pcaComp = 1
    print("Prop. of variance explained:")
    for val in pca_obj.explained_variance_ratio_:
        print("\t PCA Component " + str(pcaComp) + " = " + str(round(val, 4)))
        pcaComp = pcaComp + 1
    varExplain = pca_obj.explained_variance_ratio_

    outNComp = varExplain.shape[0]
    if n_comps is not None:
        if n_comps > varExplain.shape[0]:
            raise rsgislib.RSGISPyException(
                "You cannot output more components than "
                "the number of input image bands."
            )
        outNComp = n_comps

    pca(whiten_img, eigenVecFile, output_img, outNComp, gdalformat, datatype)
    if calc_stats:
        rsgislib.imageutils.pop_img_stats(output_img, True, in_img_no_data, True)

    if created_tmp_dir:
        shutil.rmtree(tmp_dir)
    else:
        rsgislib.tools.filetools.delete_file_silent(valid_msk_img)
        rsgislib.tools.filetools.delete_file_silent(whiten_img)
        rsgislib.tools.filetools.delete_file_silent(eigenVecFile)

    return varExplain


def rescale_img_pxl_vals(
    input_img, output_img, gdalformat, datatype, band_rescale_objs, trim_to_limits=True
):
    """
    Function which rescales an input image base on a list of rescaling parameters.

    :param input_img: the input image
    :param output_img: the output image file name and path (will be same
                       dimensions as the input)
    :param gdalformat: the GDAL image file format of the output image file.
    :param band_rescale_objs: list of ImageBandRescale objects
    :param trim_to_limits: whether to trim the output to the output min/max values.

    """
    from rios import applier

    bandRescaleDict = dict()
    for rescaleObj in band_rescale_objs:
        bandRescaleDict[rescaleObj.band - 1] = rescaleObj

    numpyDT = rsgislib.get_numpy_datatype(datatype)

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    infiles = applier.FilenameAssociations()
    infiles.image = input_img
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = output_img
    otherargs = applier.OtherInputs()
    otherargs.rescaleDict = bandRescaleDict
    otherargs.trim = trim_to_limits
    otherargs.numpyDT = numpyDT
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False

    def _applyRescale(info, inputs, outputs, otherargs):
        """
        This is an internal rios function
        """
        outputs.outimage = numpy.zeros_like(inputs.image, dtype=numpyDT)
        for idx in range(inputs.image.shape[0]):
            outputs.outimage[idx] = numpy.where(
                inputs.image[idx] == otherargs.rescaleDict[idx].no_data_val,
                otherargs.rescaleDict[idx].out_no_data,
                (
                    (
                        (inputs.image[idx] - otherargs.rescaleDict[idx].in_min)
                        / (
                            inputs.image[idx]
                            - otherargs.rescaleDict[idx].in_max
                            - inputs.image[idx]
                            - otherargs.rescaleDict[idx].in_min
                        )
                    )
                    * (
                        inputs.image[idx]
                        - otherargs.rescaleDict[idx].out_max
                        - inputs.image[idx]
                        - otherargs.rescaleDict[idx].out_min
                    )
                )
                + inputs.image[idx]
                - otherargs.rescaleDict[idx].out_min,
            )
            if otherargs.trim:
                outputs.outimage[idx] = numpy.where(
                    (outputs.outimage[idx] != otherargs.rescaleDict[idx].out_no_data)
                    & (outputs.outimage[idx] < otherargs.rescaleDict[idx].out_min),
                    otherargs.rescaleDict[idx].out_min,
                    outputs.outimage[idx],
                )
                outputs.outimage[idx] = numpy.where(
                    (outputs.outimage[idx] != otherargs.rescaleDict[idx].out_no_data)
                    & (outputs.outimage[idx] > otherargs.rescaleDict[idx].out_max),
                    otherargs.rescaleDict[idx].out_max,
                    outputs.outimage[idx],
                )

    applier.apply(_applyRescale, infiles, outfiles, otherargs, controls=aControls)


def calc_histograms_for_msk_vals(
    input_img, img_band, img_msk, msk_band, min_val, max_val, bin_width, msk_vals=None
):
    """
    A function which reads the image bands (values and mask) into memory and creates a
    histogram for each value within the mask value. Within the mask 0 is considered
    to be no data.

    :param input_img: image values image file path.
    :param img_band: values image band
    :param img_msk: file path for image mask.
    :param msk_band: mask image band
    :param min_val: minimum value for the histogram bins
    :param max_val: maximum value for the histogram bins
    :param bin_width: the width of the histograms bins.
    :param msk_vals: a list of values within the mask can be provided to just consider
                    a limited number of mask values when calculating the histograms.
                    If None (default) then calculated for all mask values.

    :return: returns a dict of mask values with an array for the histogram.

    """
    min_val = float(min_val)
    max_val = float(max_val)
    nBins = math.ceil((max_val - min_val) / bin_width)
    max_val = float(min_val + (bin_width * nBins))

    imgValsDS = gdal.Open(input_img)
    imgValsBand = imgValsDS.GetRasterBand(img_band)
    valsArr = imgValsBand.ReadAsArray()
    imgValsDS = None

    imgMskDS = gdal.Open(img_msk)
    imgMskBand = imgMskDS.GetRasterBand(msk_band)
    mskArr = imgMskBand.ReadAsArray()
    imgMskDS = None

    if msk_vals is None:
        uniq_vals = numpy.unique(mskArr)
    else:
        uniq_vals = msk_vals

    hist_dict = dict()

    for msk_val in uniq_vals:
        if msk_val != 0:
            mskd_vals = valsArr[mskArr == msk_val]
            hist_arr, bin_edges = numpy.histogram(
                mskd_vals, bins=nBins, range=(min_val, max_val)
            )
            hist_dict[msk_val] = hist_arr

    valsArr = None
    mskArr = None

    return hist_dict


def calc_imgs_pxl_mode(input_imgs, output_img, gdalformat, no_data_val=0):
    """
    Function which calculates the mode of a group of images.

    Warning, this function can be very slow!!!
    You probably want to use rsgislib.imagecalc.imagePixelColumnSummary

    :param input_imgs: the list of images
    :param output_img: the output image file name and path (will be
                       same dimensions as the input)
    :param gdalformat: the GDAL image file format of the output image file.

    """
    import rsgislib.imageutils
    import scipy.stats
    from rios import applier

    datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_imgs[0])
    numpyDT = rsgislib.get_numpy_datatype(datatype)

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    infiles = applier.FilenameAssociations()
    infiles.images = input_imgs
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = output_img
    otherargs = applier.OtherInputs()
    otherargs.no_data_val = no_data_val
    otherargs.numpyDT = numpyDT
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False

    def _applyCalcMode(info, inputs, outputs, otherargs):
        """
        This is an internal rios function
        """
        image_data = numpy.concatenate(inputs.images, axis=0).astype(numpy.float32)
        image_data[image_data == otherargs.no_data_val] = numpy.nan
        mode_arr, count_arr = scipy.stats.mode(image_data, axis=0, nan_policy="omit")
        outputs.outimage = mode_arr.astype(otherargs.numpyDT)

    applier.apply(_applyCalcMode, infiles, outfiles, otherargs, controls=aControls)


def calc_img_basic_stats_for_ref_region(
    in_ref_img, in_stats_imgs, output_img, gdalformat="KEA"
):
    """
    A function which calculates the mean and standard deviation through a series of
    input images. The region for processing is defined by the reference image and
    images padded with no-data where no data is present.

    The output image has twice the number of bands as the input image providing
    a mean and standard deviation for each input band.

    If the input images has 2 bands then the output bands will have the following
    order:

    1. band 1 mean
    2. band 1 std dev
    3. band 2 mean
    4. band 2 std dev

    :param in_ref_img: reference image which defines the output image
    :param in_stats_imgs: a list of input images over which the stats will
                          be calculated.
    :param output_img: the output image path and file name
    :param gdalformat: the output image file format. Default KEA.

    """
    import rsgislib.imageutils
    from rios import applier

    first = True
    n_bands = 0
    no_data_val = 0
    for img in in_stats_imgs:
        print(img)
        if first:
            n_bands = rsgislib.imageutils.get_img_band_count(img)
            no_data_val = rsgislib.imageutils.get_img_no_data_value(img)
            first = False
        else:
            if n_bands != rsgislib.imageutils.get_img_band_count(img):
                raise rsgislib.RSGISPyException(
                    "The number of bands must be the same in all input images."
                )
            if no_data_val != rsgislib.imageutils.get_img_no_data_value(img):
                raise rsgislib.RSGISPyException(
                    "The no data value should be the same in all input images."
                )

    # Internal function to calculate mean and standard deviation of
    # the input images
    def _calcBasicStats(info, inputs, outputs, otherargs):
        n_imgs = len(inputs.imgs)
        blk_shp = inputs.imgs[0].shape
        if blk_shp[0] != otherargs.n_bands:
            raise rsgislib.RSGISPyException(
                "Block shape and the number of input image bands do not align."
            )
        outputs.output_img = numpy.zeros(
            (blk_shp[0] * 2, blk_shp[1], blk_shp[2]), dtype=float
        )

        band_arr = []
        for band in range(blk_shp[0]):
            band_arr.append(numpy.zeros((n_imgs, blk_shp[1], blk_shp[2]), dtype=float))

        img_idx = 0
        for img_blk in inputs.imgs:
            for band in range(blk_shp[0]):
                band_arr[band][img_idx] = img_blk[band]
            img_idx = img_idx + 1

        for band in range(blk_shp[0]):
            band_arr[band][band_arr[band] == otherargs.no_data_val] = numpy.nan

            outputs.output_img[band * 2] = numpy.nanmean(band_arr[band], axis=0)
            outputs.output_img[band * 2 + 1] = numpy.nanstd(band_arr[band], axis=0)

            outputs.output_img[band * 2][
                numpy.isnan(outputs.output_img[band * 2])
            ] = otherargs.no_data_val
            outputs.output_img[band * 2 + 1][
                numpy.isnan(outputs.output_img[band * 2 + 1])
            ] = 0.0

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    infiles = applier.FilenameAssociations()
    infiles.imgs = in_stats_imgs

    otherargs = applier.OtherInputs()
    otherargs.n_bands = n_bands
    otherargs.no_data_val = no_data_val

    outfiles = applier.FilenameAssociations()
    outfiles.output_img = output_img

    aControls = applier.ApplierControls()
    aControls.referenceImage = in_ref_img
    aControls.footprint = applier.BOUNDS_FROM_REFERENCE
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Calculating Stats Image.")
    applier.apply(_calcBasicStats, infiles, outfiles, otherargs, controls=aControls)
    print("Completed")

    use_no_data = True
    if no_data_val is None:
        no_data_val = 0.0
        use_no_data = False

    rsgislib.imageutils.pop_img_stats(
        output_img, use_no_data=use_no_data, no_data_val=no_data_val, calc_pyramids=True
    )


def normalise_image_band(input_img, band, output_img, gdal_format="KEA"):
    """
    Perform a simple normalisation a single image band (val - min)/range.

    :param input_img: The input image file.
    :param band: the image band (starts at 1) to be normalised
    :param output_img: the output image file (will just be a single band).
    :param gdal_format: The output image format.

    """
    import rsgislib
    import rsgislib.imageutils

    no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img, band)
    use_no_data_val = True
    if no_data_val is None:
        use_no_data_val = False
        no_data_val = 0.0

    band_min, band_max = get_img_band_min_max(
        input_img, band, use_no_data_val, no_data_val
    )

    band_defns = [BandDefn("b1", input_img, band)]
    band_range = band_max - band_min
    exp = "(b1=={0})?0.0:(b1-{1})/{2}".format(no_data_val, band_min, band_range)
    band_math(output_img, exp, gdal_format, rsgislib.TYPE_32FLOAT, band_defns)
    rsgislib.imageutils.pop_img_stats(
        output_img, use_no_data=True, no_data_val=0.0, calc_pyramids=True
    )


def recode_int_raster(
    input_img,
    output_img,
    recode_dict,
    keep_vals_not_in_dict=True,
    gdalformat="KEA",
    datatype=rsgislib.TYPE_32INT,
):
    """
    A function recodes an input image. Assuming image only has a single image band so
    it will be band 1 which is recoded. The recode is provided as a dict where the key
    is the value to be recoded and the value of the dict is the output value.

    :param input_img: Input image file.
    :param output_img: Output image file.
    :param recode_dict: dict for recode (key: int to be recoded, value: int
                        to recode to)
    :param keep_vals_not_in_dict: boolean whether pixels not being recoded should be
                              copied to the output (True) or whether only those pixels
                              recoded should be outputted (False) (default: True)
    :param gdalformat: output file format (default: KEA)
    :param datatype: is a rsgislib.TYPE_* value providing the data type of
                     the output image.

    """
    from rios import applier

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    # Generated the combined mask.
    infiles = applier.FilenameAssociations()
    infiles.inimage = input_img
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = output_img
    otherargs = applier.OtherInputs()
    otherargs.np_dtype = rsgislib.get_numpy_datatype(datatype)
    otherargs.keep_vals_not_in_dict = keep_vals_not_in_dict
    otherargs.recode_dict = recode_dict
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = False
    aControls.calcStats = False

    def _recode(info, inputs, outputs, otherargs):
        out_arr = numpy.zeros_like(inputs.inimage[0], dtype=otherargs.np_dtype)
        if otherargs.keep_vals_not_in_dict:
            out_arr = inputs.inimage[0]

        for rc_val in recode_dict:
            out_arr[inputs.inimage[0] == rc_val] = recode_dict[rc_val]
        outputs.outimage = numpy.expand_dims(out_arr, axis=0)

    applier.apply(_recode, infiles, outfiles, otherargs, controls=aControls)


def calc_fill_regions_knn(
    in_ref_img,
    ref_no_data,
    in_fill_regions_img,
    fill_region_val,
    output_img,
    k=5,
    summary=rsgislib.SUMTYPE_MODE,
    gdalformat="KEA",
    datatype=rsgislib.TYPE_32INT,
):
    """
    A function will fills regions (defined by having a value == fill_region_val) of
    the in_fill_regions_img image using a KNN approach based of pixels within the
    in_ref_img image. The KNN distance is the spatial distance based off the
    pixel locations.

    The in_ref_img and in_fill_regions_img must both be in the same projection.

    :param in_ref_img: The reference image (can be a different resolution) from
                    which the KNN will be trained.
    :param ref_no_data: The no data value used within the reference image.
    :param in_fill_regions_img: A mask image defining the regions to be filled.
    :param fill_region_val: The pixel value specifying the pixels in the
                            in_fill_regions_img image for the KNN value will
                            be calculated.
    :param output_img: The output image file name and path.
    :param k: The k parameter of the KNN.
    :param summary: Summary method (rsgislib.SUMTYPE_*) to be used (Default is Mode).
                    Options: Mode, Median, Sum, Mean, Range, Min, Max.
    :param gdalformat: output file format (default: KEA)
    :param datatype: is a rsgislib.TYPE_* value providing the data type of the
                     output image.

    """
    import rsgislib.imageutils
    from rios import applier

    import scipy
    import scipy.stats
    import rtree

    if not rsgislib.imageutils.do_gdal_layers_have_same_proj(
        in_ref_img, in_fill_regions_img
    ):
        raise rsgislib.RSGISPyException(
            "The reference image and fill regions image "
            "do not have the same projection."
        )

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    x_res, y_res = rsgislib.imageutils.get_img_res(in_ref_img)
    if x_res < 0:
        x_res *= -1
    if y_res < 0:
        y_res *= -1

    print("Building Spatial Index...")
    # Create the index.
    kd_idx = rtree.index.Index()

    # Populate Index
    infiles = applier.FilenameAssociations()
    infiles.in_ref_img = in_ref_img
    outfiles = applier.FilenameAssociations()
    otherargs = applier.OtherInputs()
    otherargs.kd_idx = kd_idx
    otherargs.ref_no_data = ref_no_data
    otherargs.w_box = x_res / 2
    otherargs.h_box = y_res / 2
    otherargs.n = 0
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar

    def _retrieve_idx_info(info, inputs, outputs, otherargs):
        xBlock, yBlock = info.getBlockCoordArrays()
        xBlock = xBlock.flatten()
        yBlock = yBlock.flatten()
        ref_data = inputs.in_ref_img[0].flatten()

        x_coords = xBlock[ref_data != otherargs.ref_no_data]
        y_coords = yBlock[ref_data != otherargs.ref_no_data]
        val_data = ref_data[ref_data != otherargs.ref_no_data]

        left_coords = x_coords - otherargs.w_box
        right_coords = x_coords + otherargs.w_box

        top_coords = y_coords + otherargs.h_box
        bot_coords = y_coords - otherargs.h_box

        coords = numpy.stack(
            (left_coords, bot_coords, right_coords, top_coords), axis=1
        )
        for coord, cls in zip(coords, val_data):
            otherargs.kd_idx.insert(otherargs.n, coord, cls)
            otherargs.n += 1

    applier.apply(_retrieve_idx_info, infiles, outfiles, otherargs, controls=aControls)

    print("\nFilling Image Regions using KNN (k={})".format(k))
    x_res, y_res = rsgislib.imageutils.get_img_res(in_fill_regions_img)
    if x_res < 0:
        x_res *= -1
    if y_res < 0:
        y_res *= -1

    # Apply KNN fill.
    infiles = applier.FilenameAssociations()
    infiles.inimage = in_fill_regions_img
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = output_img
    otherargs = applier.OtherInputs()
    otherargs.kd_idx = kd_idx
    otherargs.k = k
    otherargs.fill_region_val = fill_region_val
    otherargs.np_dtype = rsgislib.get_numpy_datatype(datatype)
    otherargs.w_box = x_res / 2
    otherargs.h_box = y_res / 2
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = False
    aControls.calcStats = False

    def _knn_fill_regions(info, inputs, outputs, otherargs):
        out_arr = numpy.zeros_like(inputs.inimage[0], dtype=otherargs.np_dtype)
        xBlock, yBlock = info.getBlockCoordArrays()
        data_shp = out_arr.shape
        for i in range(data_shp[0]):
            for j in range(data_shp[1]):
                if inputs.inimage[0, i, j] == otherargs.fill_region_val:
                    bbox = (
                        xBlock[i, j] - otherargs.w_box,
                        yBlock[i, j] - otherargs.h_box,
                        xBlock[i, j] + otherargs.w_box,
                        yBlock[i, j] + otherargs.h_box,
                    )

                    vals = list(
                        otherargs.kd_idx.nearest(bbox, otherargs.k, objects="raw")
                    )
                    int_vals = list()
                    for val in vals:
                        int_vals.append(int(val))
                    if summary == rsgislib.SUMTYPE_MODE:
                        out_arr[i, j] = scipy.stats.mode(int_vals).mode[0]
                    elif summary == rsgislib.SUMTYPE_MEDIAN:
                        out_arr[i, j] = numpy.median(int_vals)
                    elif summary == rsgislib.SUMTYPE_MAX:
                        out_arr[i, j] = numpy.amax(int_vals)
                    elif summary == rsgislib.SUMTYPE_MIN:
                        out_arr[i, j] = numpy.amin(int_vals)
                    elif summary == rsgislib.SUMTYPE_MEAN:
                        out_arr[i, j] = numpy.mean(int_vals)
                    elif summary == rsgislib.SUMTYPE_SUM:
                        out_arr[i, j] = numpy.sum(int_vals)
                    elif summary == rsgislib.SUMTYPE_RANGE:
                        out_arr[i, j] = numpy.amax(int_vals) - numpy.amin(int_vals)
                    else:
                        raise rsgislib.RSGISPyException(
                            "Summary method not recognised/available."
                        )

        outputs.outimage = numpy.expand_dims(out_arr, axis=0)

    applier.apply(_knn_fill_regions, infiles, outfiles, otherargs, controls=aControls)
    print("Finished fill")


def are_imgs_equal(in_ref_img, in_cmp_img, prop_eql=1.0, flt_dif=0.0001):
    """
    A function to check whether two images have equal pixel values within the spatial
    overlap. Note, if the two input images only have a partial overlap (i.e., one is
    a subset of the other) then they are still be equal if the overlapping region
    has matching pixel values.

    :param in_ref_img: The input reference image for the comparison.
    :param in_cmp_img: The input comparison image to be compared to the reference image.
    :param prop_eql: The proportion of pixels within the scene which need to be
                     identified as identical to return True. Range is 0 - 1. Default
                     is 1.0 (i.e., 100 % of pixels have to be identical to return True).
    :param flt_dif: A threshold for comparing two floating point numbers as being
                    identical - this avoids issues with rounding and the number of
                    decimal figures stored.
    :return: Boolean (match), float (proportion of pixels which matched)

    """
    import rsgislib.imageutils
    from rios import applier

    if rsgislib.imageutils.get_img_band_count(
        in_ref_img
    ) != rsgislib.imageutils.get_img_band_count(in_cmp_img):
        raise rsgislib.RSGISPyException(
            "The number of image bands is not the same between the two images."
        )

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    # Generated the combined mask.
    infiles = applier.FilenameAssociations()
    infiles.in_ref_img = in_ref_img
    infiles.in_cmp_img = in_cmp_img
    outfiles = applier.FilenameAssociations()
    otherargs = applier.OtherInputs()
    otherargs.flt_dif = flt_dif
    otherargs.n_pxls = 0.0
    otherargs.n_eq_pxls = 0.0
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar

    def _calcPropEqual(info, inputs, outputs, otherargs):
        ref_pxls = inputs.in_ref_img.flatten()
        cmp_pxls = inputs.in_cmp_img.flatten()

        cmp_pxls = cmp_pxls[~numpy.isnan(ref_pxls)]
        ref_pxls = ref_pxls[~numpy.isnan(ref_pxls)]

        ref_pxls = ref_pxls[~numpy.isnan(cmp_pxls)]
        cmp_pxls = cmp_pxls[~numpy.isnan(cmp_pxls)]

        pxl_diff = numpy.abs(ref_pxls - cmp_pxls)
        eql_pxls = numpy.where(pxl_diff < otherargs.flt_dif, 1, 0)

        otherargs.n_eq_pxls += eql_pxls.sum()
        otherargs.n_pxls += ref_pxls.shape[0]

    applier.apply(_calcPropEqual, infiles, outfiles, otherargs, controls=aControls)

    prop_pxls_eq = otherargs.n_eq_pxls / otherargs.n_pxls
    return (prop_pxls_eq >= prop_eql), prop_pxls_eq


def are_img_bands_equal(
    in_ref_img, img_ref_band, in_cmp_img, img_cmp_band, prop_eql=1.0, flt_dif=0.0001
):
    """
    A function to check whether two image bands have equal pixel values within the
    spatial overlap. Note, if the two input images only have a partial overlap
    (i.e., one is a subset of the other) then they are still be equal if the
    overlapping region has matching pixel values.

    :param in_ref_img: The input reference image for the comparison.
    :param img_ref_band: The band from the reference image
    :param in_cmp_img: The input comparison image to be compared to the reference image.
    :param img_cmp_band: The band from the comparison image
    :param prop_eql: The proportion of pixels within the scene which need to be
                     identified as identical to return True. Range is 0 - 1. Default
                     is 1.0 (i.e., 100 % of pixels have to be identical to return True).
    :param flt_dif: A threshold for comparing two floating point numbers as being
                    identical - this avoids issues with rounding and the number of
                    decimal figures stored.
    :return: Boolean (match), float (proportion of pixels which matched)

    """
    import rsgislib.imageutils
    from rios import applier

    n_ref_bands = rsgislib.imageutils.get_img_band_count(in_ref_img)
    n_cmp_bands = rsgislib.imageutils.get_img_band_count(in_cmp_img)

    if (img_ref_band < 1) or (img_ref_band > n_ref_bands):
        raise rsgislib.RSGISPyException(
            "The specified band is not within the reference image."
        )

    if (img_cmp_band < 1) or (img_cmp_band > n_cmp_bands):
        raise rsgislib.RSGISPyException(
            "The specified band is not within the comparison image."
        )

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    # Generated the combined mask.
    infiles = applier.FilenameAssociations()
    infiles.in_ref_img = in_ref_img
    infiles.in_cmp_img = in_cmp_img
    outfiles = applier.FilenameAssociations()
    otherargs = applier.OtherInputs()
    otherargs.ref_band_idx = img_ref_band - 1
    otherargs.cmp_band_idx = img_cmp_band - 1
    otherargs.flt_dif = flt_dif
    otherargs.n_pxls = 0.0
    otherargs.n_eq_pxls = 0.0
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar

    def _calcBandPropEqual(info, inputs, outputs, otherargs):
        ref_pxls = inputs.in_ref_img[otherargs.ref_band_idx, ...].flatten()
        cmp_pxls = inputs.in_cmp_img[otherargs.cmp_band_idx, ...].flatten()

        cmp_pxls = cmp_pxls[~numpy.isnan(ref_pxls)]
        ref_pxls = ref_pxls[~numpy.isnan(ref_pxls)]

        ref_pxls = ref_pxls[~numpy.isnan(cmp_pxls)]
        cmp_pxls = cmp_pxls[~numpy.isnan(cmp_pxls)]

        pxl_diff = numpy.abs(ref_pxls - cmp_pxls)
        eql_pxls = numpy.where(pxl_diff < otherargs.flt_dif, 1, 0)

        otherargs.n_eq_pxls += eql_pxls.sum()
        otherargs.n_pxls += ref_pxls.shape[0]

    applier.apply(_calcBandPropEqual, infiles, outfiles, otherargs, controls=aControls)

    prop_pxls_eq = otherargs.n_eq_pxls / otherargs.n_pxls
    return (prop_pxls_eq >= prop_eql), prop_pxls_eq



import os
import pytest
from shutil import copy2

GEOPANDAS_NOT_AVAIL = False
try:
    import geopandas
except ImportError:
    GEOPANDAS_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
IMGUTILS_DATA_DIR = os.path.join(DATA_DIR, "imageutils")
IMGREG_DATA_DIR = os.path.join(DATA_DIR, "imageregistration")


def test_set_env_vars_lzw_gtiff_outs():
    import rsgislib.imageutils

    rsgislib.imageutils.set_env_vars_lzw_gtiff_outs(bigtiff=False)
    assert os.environ["RSGISLIB_IMG_CRT_OPTS_GTIFF"] == "TILED=YES:COMPRESS=LZW"


def test_set_env_vars_lzw_gtiff_outs_bigtiff():
    import rsgislib.imageutils

    rsgislib.imageutils.set_env_vars_lzw_gtiff_outs(bigtiff=True)
    assert (
        os.environ["RSGISLIB_IMG_CRT_OPTS_GTIFF"]
        == "TILED=YES:COMPRESS=LZW:BIGTIFF=YES"
    )


def test_set_env_vars_deflate_gtiff_outs():
    import rsgislib.imageutils

    rsgislib.imageutils.set_env_vars_deflate_gtiff_outs(bigtiff=False)
    assert os.environ["RSGISLIB_IMG_CRT_OPTS_GTIFF"] == "TILED=YES:COMPRESS=DEFLATE"


def test_set_env_vars_deflate_gtiff_outs_bigtiff():
    import rsgislib.imageutils

    rsgislib.imageutils.set_env_vars_deflate_gtiff_outs(bigtiff=True)
    assert (
        os.environ["RSGISLIB_IMG_CRT_OPTS_GTIFF"]
        == "TILED=YES:COMPRESS=DEFLATE:BIGTIFF=YES"
    )


def test_get_gdal_img_creation_opts_none():
    import rsgislib.imageutils

    create_opts = rsgislib.imageutils.get_gdal_img_creation_opts(gdalformat="KEA")
    assert create_opts is None


def test_get_gdal_img_creation_opts_gtiff():
    import rsgislib.imageutils

    rsgislib.imageutils.set_env_vars_lzw_gtiff_outs(bigtiff=False)
    create_opts = rsgislib.imageutils.get_gdal_img_creation_opts(gdalformat="GTIFF")
    assert (create_opts["TILED"] == "YES") and (create_opts["COMPRESS"] == "LZW")


def test_get_rsgislib_datatype_from_img():
    import rsgislib
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    assert (
        rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        == rsgislib.TYPE_16UINT
    )


def test_get_gdal_datatype_from_img():
    from osgeo import gdal
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    assert rsgislib.imageutils.get_gdal_datatype_from_img(input_img) == gdal.GDT_UInt16


def test_get_gdal_datatype_name_from_img():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    assert rsgislib.imageutils.get_gdal_datatype_name_from_img(input_img) == "UInt16"


def test_get_image_res():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    x_res, y_res = rsgislib.imageutils.get_img_res(input_img)
    assert (x_res == 10) and (y_res == -10)


def test_get_image_res_abs():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    x_res, y_res = rsgislib.imageutils.get_img_res(input_img, abs_vals=True)
    assert (x_res == 10) and (y_res == 10)


def get_img_size():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    x_size, y_size = rsgislib.imageutils.get_img_size(input_img)
    assert (x_size == 931) and (y_size == 947)


def test_get_image_band_count():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    assert rsgislib.imageutils.get_img_band_count(input_img) == 10


def test_get_image_bbox():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    bbox = rsgislib.imageutils.get_img_bbox(input_img)
    assert (
        (abs(bbox[0] - 257060.59483340546) < 0.0001)
        and (abs(bbox[1] - 266370.59483340546) < 0.0001)
        and (abs(bbox[2] - 276258.7167059921) < 0.0001)
        and (abs(bbox[3] - 285728.7167059921) < 0.0001)
    )


def test_get_img_bbox_in_proj_4326():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    bbox = rsgislib.imageutils.get_img_bbox_in_proj(input_img, out_epsg=4326)
    assert (
        (abs(bbox[0] - -4.105) < 0.1)
        and (abs(bbox[1] - -3.964) < 0.1)
        and (abs(bbox[2] - 52.366) < 0.1)
        and (abs(bbox[3] - 52.453) < 0.1)
    )


def test_get_img_bbox_in_proj_utm30n():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    bbox = rsgislib.imageutils.get_img_bbox_in_proj(input_img, out_epsg=32630)
    assert (
        (abs(bbox[0] - 424918.88) < 3.0)
        and (abs(bbox[1] - 434357.09) < 3.0)
        and (abs(bbox[2] - 5802283.59) < 3.0)
        and (abs(bbox[3] - 5811879.561707063) < 3.0)
    )


def test_get_img_subset_pxl_bbox():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    sub_bbox = [260000, 264000, 280000, 285000]
    bbox = rsgislib.imageutils.get_img_subset_pxl_bbox(input_img, sub_bbox=sub_bbox)
    assert (
        (bbox[0] == 294) and (bbox[1] == 694) and (bbox[2] == 73) and (bbox[3] == 573)
    )


def test_get_img_pxl_spatial_coords():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    sub_pxl_bbox = [294, 694, 73, 573]
    bbox = rsgislib.imageutils.get_img_pxl_spatial_coords(
        input_img, sub_pxl_bbox=sub_pxl_bbox
    )
    assert (
        (abs(bbox[0] - 260000.59) < 1.0)
        and (abs(bbox[1] - 264000.59) < 1.0)
        and (abs(bbox[2] - 279998.72) < 1.0)
        and (abs(bbox[3] - 284998.72) < 1.0)
    )


def test_subset_roi_contained(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(IMGUTILS_DATA_DIR, "roi_subset.geojson")
    vec_lyr = "roi_subset"
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")

    rsgislib.imageutils.subset(
        input_img, vec_file, vec_lyr, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    n_bands = rsgislib.imageutils.get_img_band_count(output_img)
    if n_bands != 10:
        assert False

    x_res, y_res = rsgislib.imageutils.get_img_res(output_img)
    if (x_res != 10) and (y_res != 10):
        assert False

    x_size, y_size = rsgislib.imageutils.get_img_size(output_img)
    if (x_size != 325) and (y_size != 340):
        assert False

    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(input_img, output_img)
    assert img_eq


def test_subset_roi_overlap(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(IMGUTILS_DATA_DIR, "roi_subset_part_overlap.geojson")
    vec_lyr = "roi_subset_part_overlap"
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")

    rsgislib.imageutils.subset(
        input_img, vec_file, vec_lyr, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    n_bands = rsgislib.imageutils.get_img_band_count(output_img)
    if n_bands != 10:
        assert False

    x_res, y_res = rsgislib.imageutils.get_img_res(output_img)
    if (x_res != 10) and (y_res != 10):
        assert False

    x_size, y_size = rsgislib.imageutils.get_img_size(output_img)
    if (x_size != 931) and (y_size != 238):
        assert False

    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(input_img, output_img)
    assert img_eq


def test_pop_img_stats(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.pop_img_stats(
        input_img, use_no_data=True, no_data_val=0, calc_pyramids=True
    )


def test_pop_thmt_img_stats(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.pop_thmt_img_stats(
        input_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
    )


def test_get_img_band_stats():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")

    min, max, mean, std = rsgislib.imageutils.get_img_band_stats(
        input_img, img_band=1, compute=True
    )

    assert (
        (min == 1.0)
        and (max == 1066.0)
        and (abs(mean - 63.13) < 0.01)
        and (abs(std - 32.44) < 0.01)
    )


def test_get_img_no_data_value_zero():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")

    no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)

    assert no_data_val == 0


def test_get_img_no_data_value_none():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")

    no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)

    assert no_data_val is None


def test_set_img_no_data_value(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_no_data_value(input_img, no_data_val=100)
    no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)

    assert no_data_val == 100


def test_get_img_files():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    files = rsgislib.imageutils.get_img_files(input_img)
    assert len(files) == 1


def test_has_gcps_false():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    assert not rsgislib.imageutils.has_gcps(input_img)


def test_has_gcps_true():
    import rsgislib.imageutils

    input_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset_gcps.kea"
    )
    assert rsgislib.imageutils.has_gcps(input_img)


def test_is_img_thematic_False():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    assert not rsgislib.imageutils.is_img_thematic(input_img)


def test_is_img_thematic_True():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    assert rsgislib.imageutils.is_img_thematic(input_img)


def test_get_gdal_format_name():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    assert "KEA" == rsgislib.imageutils.get_gdal_format_name(input_img)


def test_get_epsg_proj_from_img_27700():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    assert 27700 == rsgislib.imageutils.get_epsg_proj_from_img(input_img)


def test_get_epsg_proj_from_img_32630():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    assert 32630 == rsgislib.imageutils.get_epsg_proj_from_img(input_img)


def test_get_wkt_proj_from_img_osgb():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    assert "British National Grid" in rsgislib.imageutils.get_wkt_proj_from_img(
        input_img
    )


def test_get_wkt_proj_from_img_utm():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    assert "UTM zone 30N" in rsgislib.imageutils.get_wkt_proj_from_img(input_img)


def test_assign_wkt_proj_file(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    wkt_file = os.path.join(IMGUTILS_DATA_DIR, "utm30n.wkt")

    rsgislib.imageutils.assign_wkt_proj(input_img, None, wkt_file)

    assert "UTM zone 30N" in rsgislib.imageutils.get_wkt_proj_from_img(input_img)


def test_assign_wkt_proj_str(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    wkt_str = """PROJCS["WGS 84 / UTM zone 30N",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",-3],PARAMETER["scale_factor",0.9996],PARAMETER["false_easting",500000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","32630"]]"""

    rsgislib.imageutils.assign_wkt_proj(input_img, wkt_str, None)

    assert "UTM zone 30N" in rsgislib.imageutils.get_wkt_proj_from_img(input_img)


def test_assign_spatial_info(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.assign_spatial_info(
        input_img, None, None, 100, -100, None, None
    )
    x_res, y_res = rsgislib.imageutils.get_img_res(input_img)
    assert (x_res == 100) and (y_res == -100)


def test_copy_proj_from_img(tmp_path):
    import rsgislib.imageutils

    input_utm_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.copy_proj_from_img(input_img, input_utm_img)

    assert "UTM zone 30N" in rsgislib.imageutils.get_wkt_proj_from_img(input_img)


def test_copy_spatial_and_proj_from_img(tmp_path):
    import rsgislib.imageutils

    input_utm_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.copy_spatial_and_proj_from_img(input_img, input_utm_img)

    assert "UTM zone 30N" in rsgislib.imageutils.get_wkt_proj_from_img(input_img)


def test_get_utm_zone():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    assert "30N" == rsgislib.imageutils.get_utm_zone(input_img)


def test_copy_gcps(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset_gcps.kea"
    )

    output_ref_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset.kea"
    )
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_subset_b123_offset.kea")
    copy2(output_ref_img, output_img)

    rsgislib.imageutils.copy_gcps(input_img, output_img)

    assert rsgislib.imageutils.has_gcps(output_img)


def test_resample_img_to_match(tmp_path):
    import rsgislib.imageutils

    input_utm_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    input_osgb_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.resample_img_to_match(
        input_utm_img, input_osgb_img, output_img, "KEA"
    )

    assert os.path.exists(output_img)


def test_reproject_image(tmp_path):
    import rsgislib.imageutils

    input_osgb_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    wkt_file = os.path.join(IMGUTILS_DATA_DIR, "utm30n.wkt")

    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.reproject_image(input_osgb_img, output_img, wkt_file)

    assert os.path.exists(output_img)


def test_gdal_warp(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.gdal_warp(
        input_img, output_img, 4326, gdalformat="KEA", use_multi_threaded=False
    )

    assert os.path.exists(output_img)


def test_create_img_mosaic(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_img_mosaic(
        imgs, output_img, 0, 0, 1, 0, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_include_imgs(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img(
        input_ref_img, output_img, 10, 0, "KEA", rsgislib.TYPE_16UINT
    )

    imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    rsgislib.imageutils.include_imgs(output_img, imgs)

    assert os.path.exists(output_img)


def test_include_imgs_with_overlap(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img(
        input_ref_img, output_img, 10, 0, "KEA", rsgislib.TYPE_16UINT
    )

    imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_over_tiles", "*.kea"))
    rsgislib.imageutils.include_imgs_with_overlap(output_img, imgs, overlap=20)

    assert os.path.exists(output_img)


def test_include_imgs_ind_img_intersect(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img(
        input_ref_img, output_img, 10, 0, "KEA", rsgislib.TYPE_16UINT
    )

    imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    rsgislib.imageutils.include_imgs_ind_img_intersect(output_img, imgs)

    assert os.path.exists(output_img)


def test_combine_imgs_to_band(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "cls_imgs", "*.kea"))
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.combine_imgs_to_band(
        imgs, output_img, gdalformat="KEA", datatype=rsgislib.TYPE_8UINT, no_data_val=0
    )

    assert os.path.exists(output_img)


# TODO rsgislib.imageutils.create_ref_img_composite_img


def test_combine_binary_masks(tmp_path):
    import rsgislib
    import rsgislib.imageutils

    msk_imgs_dict = dict()
    msk_imgs_dict["forest"] = os.path.join(
        IMGUTILS_DATA_DIR, "cls_imgs", "Forest_1226565d33.kea"
    )
    msk_imgs_dict["grass"] = os.path.join(
        IMGUTILS_DATA_DIR, "cls_imgs", "Grass_1226565d33.kea"
    )
    msk_imgs_dict["urban"] = os.path.join(
        IMGUTILS_DATA_DIR, "cls_imgs", "Urban_1226565d33.kea"
    )
    msk_imgs_dict["water"] = os.path.join(
        IMGUTILS_DATA_DIR, "cls_imgs", "Water_1226565d33.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    out_lut_file = os.path.join(tmp_path, "out_lut.json")
    rsgislib.imageutils.combine_binary_masks(
        msk_imgs_dict, output_img, out_lut_file, gdalformat="KEA"
    )

    assert os.path.exists(output_img) and os.path.exists(out_lut_file)


# TODO rsgislib.imageutils.export_single_merged_img_band
# TODO rsgislib.imageutils.order_img_using_prop_valid_pxls
# TODO rsgislib.imageutils.gen_timeseries_fill_composite_img


def test_create_tiles(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_img_base = os.path.join(tmp_path, "out_img")
    rsgislib.imageutils.create_tiles(
        input_img, out_img_base, 200, 200, 0, False, "KEA", rsgislib.TYPE_16UINT, "kea"
    )

    assert len(glob.glob("{}*.kea".format(out_img_base))) == 25


def test_create_tiles_outpath_exp(tmp_path):
    import rsgislib
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_img_base = os.path.join(tmp_path, "output1", "output2", "out_img")
    with pytest.raises(Exception):
        rsgislib.imageutils.create_tiles(
            input_img,
            out_img_base,
            200,
            200,
            0,
            False,
            "KEA",
            rsgislib.TYPE_16UINT,
            "kea",
        )


def test_create_tiles_multi_core(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_img_base = os.path.join(tmp_path, "out_img")
    rsgislib.imageutils.create_tiles_multi_core(
        input_img, out_img_base, 200, 200, "KEA", rsgislib.TYPE_16UINT, "kea", 1
    )

    assert len(glob.glob("{}*.kea".format(out_img_base))) == 25


def test_create_tiles_multi_core_outpath_exp(tmp_path):
    import rsgislib
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_img_base = os.path.join(tmp_path, "output1", "output2", "out_img")
    print(out_img_base)
    with pytest.raises(Exception):
        rsgislib.imageutils.create_tiles_multi_core(
            input_img, out_img_base, 200, 200, "KEA", rsgislib.TYPE_16UINT, "kea", 1
        )


def test_stretch_img(tmp_path):
    import rsgislib
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.stretch_img(
        input_img,
        output_img,
        False,
        "",
        0,
        False,
        "KEA",
        rsgislib.TYPE_8UINT,
        rsgislib.imageutils.STRETCH_LINEARPERCENT,
        2,
    )

    assert os.path.exists(output_img)


def test_stretch_img_with_stats(tmp_path):
    import rsgislib
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    in_stats_file = os.path.join(IMGUTILS_DATA_DIR, "s2_stch_stats.txt")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.stretch_img_with_stats(
        input_img,
        output_img,
        in_stats_file,
        "KEA",
        rsgislib.TYPE_8UINT,
        0,
        rsgislib.imageutils.STRETCH_LINEARMINMAX,
        2,
    )

    assert os.path.exists(output_img)


def test_normalise_img_pxl_vals(tmp_path):
    import rsgislib
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.normalise_img_pxl_vals(
        input_img,
        output_img,
        "KEA",
        rsgislib.TYPE_32FLOAT,
        in_no_data_val=0,
        out_no_data_val=0,
        out_min=0,
        out_max=1,
        stretch_type=rsgislib.imageutils.STRETCH_LINEARPERCENT,
        stretch_param=2,
    )

    assert os.path.exists(output_img)


def test_get_img_band_colour_interp():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")

    assert rsgislib.imageutils.get_img_band_colour_interp(input_img, 1) == 1


def test_set_img_band_colour_interp(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_band_colour_interp(
        input_img, img_band=1, clr_itrp_val=3
    )

    assert rsgislib.imageutils.get_img_band_colour_interp(input_img, 1) == 3


def test_set_img_thematic(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_thematic(input_img)

    assert rsgislib.imageutils.is_img_thematic(input_img)


def test_set_img_not_thematic(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_not_thematic(input_img)

    assert not rsgislib.imageutils.is_img_thematic(input_img)


def test_define_colour_table(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, input_img)

    clr_lut = dict()
    clr_lut[1] = "#009600"
    clr_lut[2] = "#FFE5CC"
    clr_lut[3] = "#CCFFE5"
    rsgislib.imageutils.define_colour_table(input_img, clr_lut)


def test_mask_img(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.mask_img(
        input_img, in_msk_img, output_img, "KEA", rsgislib.TYPE_16UINT, 0, 0
    )

    assert os.path.exists(output_img)


def test_gen_finite_mask(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.gen_finite_mask(input_img, output_img, gdalformat="KEA")

    assert os.path.exists(output_img)


def test_gen_valid_mask(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.gen_valid_mask(
        input_img, output_img, gdalformat="KEA", no_data_val=0.0
    )

    assert os.path.exists(output_img)


def test_gen_img_edge_mask(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.gen_img_edge_mask(
        input_img, output_img, gdalformat="KEA", n_edge_pxls=10
    )

    assert os.path.exists(output_img)


def test_mask_img_with_vec(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    roi_vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    roi_vec_lyr = "aber_osgb_single_poly"
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.mask_img_with_vec(
        input_img,
        output_img,
        "KEA",
        roi_vec_file,
        roi_vec_lyr,
        tmp_path,
        outvalue=0,
        datatype=rsgislib.TYPE_16UINT,
        vec_epsg=None,
    )

    assert os.path.exists(output_img)


def test_mask_all_band_zero_vals(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.mask_all_band_zero_vals(
        input_img, output_img, gdalformat="KEA", out_val=1
    )

    assert os.path.exists(output_img)


def test_create_valid_mask(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber.kea")
    copy2(input_ref_img, input_img)
    rsgislib.imageutils.set_img_no_data_value(input_img, 0.0)

    img_band_info = list()
    img_band_info.append(
        rsgislib.imageutils.ImageBandInfo(
            file_name=input_img, name="s2_1", bands=[1, 2]
        )
    )
    img_band_info.append(
        rsgislib.imageutils.ImageBandInfo(
            file_name=input_img, name="s2_2", bands=[3, 4]
        )
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_valid_mask(
        img_band_info, output_img, gdalformat="KEA", tmp_dir=tmp_path
    )

    assert os.path.exists(output_img)


def test_subset(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_lyr = "aber_osgb_single_poly"
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.subset(
        input_img, vec_file, vec_lyr, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_subset_to_img(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_roi_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.subset_to_img(
        input_img, in_roi_img, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_subset_to_geoms_bbox(tmp_path):
    import rsgislib.imageutils
    import glob

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"
    out_img_base = os.path.join(tmp_path, "out_img")
    rsgislib.imageutils.subset_to_geoms_bbox(
        input_img,
        vec_file,
        vec_lyr,
        "val",
        out_img_base,
        gdalformat="KEA",
        datatype=None,
        out_img_ext="kea",
    )

    assert len(glob.glob("{}*.kea".format(out_img_base))) == 4


def test_subset_bbox(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_roi_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    bbox = rsgislib.imageutils.get_img_bbox(in_roi_img)
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.subset_bbox(
        input_img,
        output_img,
        "KEA",
        rsgislib.TYPE_16UINT,
        bbox[0],
        bbox[1],
        bbox[2],
        bbox[3],
    )

    assert os.path.exists(output_img)


def test_subset_pxl_bbox(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.subset_pxl_bbox(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT, 200, 300, 200, 300
    )

    assert os.path.exists(output_img)


def test_subset_to_vec(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    roi_vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    roi_vec_lyr = "aber_osgb_single_poly"
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.subset_to_vec(
        input_img,
        output_img,
        "KEA",
        roi_vec_file,
        roi_vec_lyr,
        datatype=rsgislib.TYPE_16UINT,
        vec_epsg=27700,
    )

    assert os.path.exists(output_img)


def test_perform_random_pxl_sample_in_mask(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "aber_osgb_multi_polys_rasters.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.perform_random_pxl_sample_in_mask(
        input_img, output_img, gdalformat="KEA", mask_vals=1, n_samples=10
    )
    assert os.path.exists(output_img)


def test_perform_random_pxl_sample_in_mask_low_pxl_count(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "aber_osgb_multi_polys_rasters.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.perform_random_pxl_sample_in_mask_low_pxl_count(
        input_img, output_img, gdalformat="KEA", mask_vals=1, n_samples=10
    )
    assert os.path.exists(output_img)


def test_extract_img_pxl_sample():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "aber_osgb_multi_polys_rasters.kea")
    vals = rsgislib.imageutils.extract_img_pxl_sample(input_img, pxl_n_sample=10)
    assert vals.shape[0] == 88165


def test_extract_img_pxl_vals_in_msk():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_msk_img = os.path.join(DATA_DIR, "aber_osgb_multi_polys_rasters.kea")
    vals = rsgislib.imageutils.extract_img_pxl_vals_in_msk(
        input_img, img_bands=[1, 2, 3], in_msk_img=in_msk_img, img_mask_val=4
    )
    assert vals.shape[0] == 21957


def test_create_blank_img(tmp_path):
    import rsgislib.imageutils

    wkt_file = os.path.join(IMGUTILS_DATA_DIR, "utm30n.wkt")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_blank_img(
        output_img,
        1,
        200,
        200,
        20000,
        500000,
        50,
        -100,
        0,
        wkt_file,
        "",
        "KEA",
        rsgislib.TYPE_32INT,
    )
    assert os.path.exists(output_img)


def test_create_blank_img_py(tmp_path):
    import rsgislib.imageutils
    import rsgislib.tools.utils

    wkt_file = os.path.join(IMGUTILS_DATA_DIR, "utm30n.wkt")
    wkt_str = rsgislib.tools.utils.read_text_file_no_new_lines(wkt_file)
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_blank_img_py(
        output_img,
        3,
        500,
        500,
        20000,
        500000,
        50,
        -10,
        wkt_str,
        "KEA",
        rsgislib.TYPE_32UINT,
    )
    assert os.path.exists(output_img)


def test_create_copy_img(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 3, 5, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)


def test_create_copy_img_def_extent(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    bbox = rsgislib.imageutils.get_img_bbox(input_img)
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img_def_extent(
        input_img,
        output_img,
        2,
        1,
        bbox[0],
        bbox[1],
        bbox[2],
        bbox[3],
        20,
        -20,
        "KEA",
        rsgislib.TYPE_32FLOAT,
    )
    assert os.path.exists(output_img)


def test_create_blank_buf_img_from_ref_img(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_blank_buf_img_from_ref_img(
        input_img, output_img, "KEA", rsgislib.TYPE_32FLOAT, 50, no_data_val=0
    )
    assert os.path.exists(output_img)


def test_create_blank_img_from_ref_vector(tmp_path):
    import rsgislib.imageutils

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_lyr = "aber_osgb_single_poly"
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_blank_img_from_ref_vector(
        vec_file, vec_lyr, output_img, 20, 4, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)


def test_create_copy_img_vec_extent_snap_to_grid(tmp_path):
    import rsgislib.imageutils

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_lyr = "aber_osgb_single_poly"
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img_vec_extent_snap_to_grid(
        vec_file, vec_lyr, output_img, 20, 4, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)


def test_create_blank_img_from_bbox(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    bbox = rsgislib.imageutils.get_img_bbox(input_img)
    wkt_file = os.path.join(IMGUTILS_DATA_DIR, "utm30n.wkt")
    wkt_str = rsgislib.tools.utils.read_text_file_no_new_lines(wkt_file)
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_blank_img_from_bbox(
        bbox, wkt_str, output_img, 5, 10, 5, "KEA", rsgislib.TYPE_32FLOAT, True
    )
    assert os.path.exists(output_img)


def test_create_img_for_each_vec_feat(tmp_path):
    import rsgislib.imageutils
    import glob

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"
    rsgislib.imageutils.create_img_for_each_vec_feat(
        vec_file,
        vec_lyr,
        file_name_col="tile_name",
        out_img_path=tmp_path,
        out_img_ext="kea",
        out_img_pxl_val=0,
        out_img_n_bands=1,
        out_img_res=25,
        gdalformat="KEA",
        datatype=rsgislib.TYPE_16INT,
        snap_to_grid=True,
    )
    assert len(glob.glob(os.path.join(tmp_path, "*.kea"))) == 4


def test_create_copy_img_vec_extent(tmp_path):
    import rsgislib.imageutils

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_lyr = "aber_osgb_single_poly"
    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img_vec_extent(
        input_img, vec_file, vec_lyr, output_img, 1, 10, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)


def test_gdal_translate(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.tif")
    rsgislib.imageutils.gdal_translate(input_img, output_img, gdalformat="GTIFF")
    assert os.path.exists(output_img)


def test_create_stack_images_vrt(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.vrt")
    rsgislib.imageutils.create_stack_images_vrt([input_img, input_img], output_img)
    assert os.path.exists(output_img)


def test_create_mosaic_images_vrt(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    output_img = os.path.join(tmp_path, "out_img.vrt")
    rsgislib.imageutils.create_mosaic_images_vrt(imgs, output_img)

    assert os.path.exists(output_img)


def test_create_mosaic_images_vrt_extent(tmp_path):
    import rsgislib
    import rsgislib.imageutils
    import glob

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    bbox = rsgislib.imageutils.get_img_bbox(input_img)

    imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    output_img = os.path.join(tmp_path, "out_img.vrt")
    rsgislib.imageutils.create_mosaic_images_vrt(imgs, output_img, vrt_extent=bbox)

    assert os.path.exists(output_img)


def test_create_vrt_band_subset(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.vrt")
    rsgislib.imageutils.create_vrt_band_subset(input_img, [1, 2, 3], output_img)
    assert os.path.exists(output_img)


def test_select_img_bands(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.select_img_bands(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT, [1, 2, 3]
    )
    assert os.path.exists(output_img)


def test_stack_img_bands(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.stack_img_bands(
        [input_img, input_img], None, output_img, 0, 0, "KEA", rsgislib.TYPE_16UINT
    )
    assert os.path.exists(output_img)


# TODO rsgislib.imageutils.pan_sharpen_hcs
# TODO rsgislib.imageutils.sharpen_low_res_bands


def test_get_band_names():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    band_names = rsgislib.imageutils.get_band_names(input_img)
    assert len(band_names) == 3


def test_set_band_names(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset_b123.kea")
    copy2(input_ref_img, input_img)
    rsgislib.imageutils.set_band_names(input_img, band_names=["b1", "b2", "b3"])

    band_names = rsgislib.imageutils.get_band_names(input_img)
    assert band_names[0] == "b1" and band_names[1] == "b2" and band_names[2] == "b3"


def test_do_img_res_match():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    assert rsgislib.imageutils.do_img_res_match(input_img, input_img)


def test_check_img_lst():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    input_imgs = [input_img, input_img, input_img]
    bbox = rsgislib.imageutils.get_img_bbox(input_img)
    rsgislib.imageutils.check_img_lst(
        input_imgs, 10, -10, bbox, print_errors=True, abs_res=True
    )


def test_check_img_file_comparison():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    rsgislib.imageutils.check_img_file_comparison(
        input_img, input_img, test_n_bands=True, test_eql_bbox=True, print_errors=True
    )


def test_do_images_overlap():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    input_sub_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")

    assert rsgislib.imageutils.do_images_overlap(input_img, input_sub_img)


def test_do_gdal_layers_have_same_proj_true():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    input_sub_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")

    assert rsgislib.imageutils.do_gdal_layers_have_same_proj(input_img, input_sub_img)


def test_do_gdal_layers_have_same_proj_false():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    input_utm_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")

    assert not rsgislib.imageutils.do_gdal_layers_have_same_proj(
        input_img, input_utm_img
    )


def test_test_img_lst_intersects():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    input_sub_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    rsgislib.imageutils.test_img_lst_intersects([input_img, input_sub_img])


def test_calc_wgs84_pixel_area(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_wgs84.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.imageutils.calc_wgs84_pixel_area(input_img, output_img, gdalformat="KEA")

    assert os.path.exists(output_img)


def test_calc_wsg84_pixel_size(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_wgs84.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.imageutils.calc_wsg84_pixel_size(input_img, output_img, gdalformat="KEA")

    assert os.path.exists(output_img)


def test_calc_pixel_locations(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.imageutils.calc_pixel_locations(input_img, output_img, gdalformat="KEA")

    assert os.path.exists(output_img)


def test_get_file_img_extension_kea():
    import rsgislib.imageutils

    assert rsgislib.imageutils.get_file_img_extension("KEA") == "kea"


def test_get_file_img_extension_tif():
    import rsgislib.imageutils

    assert rsgislib.imageutils.get_file_img_extension("GTIFF") == "tif"


def test_get_file_img_extension_hfa():
    import rsgislib.imageutils

    assert rsgislib.imageutils.get_file_img_extension("HFA") == "img"


def test_get_file_img_extension_envi():
    import rsgislib.imageutils

    assert rsgislib.imageutils.get_file_img_extension("ENVI") == "env"


def test_get_file_img_extension_pcidsk():
    import rsgislib.imageutils

    assert rsgislib.imageutils.get_file_img_extension("PCIDSK") == "pix"


def test_rename_gdal_layer(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.rename_gdal_layer(input_img, output_img)

    assert os.path.exists(output_img)


def test_delete_gdal_layer(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.delete_gdal_layer(input_img)

    assert not os.path.exists(input_img)


def test_get_img_pxl_values():
    import rsgislib.imageutils
    import numpy

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    pxl_vals = rsgislib.imageutils.get_img_pxl_values(
        input_img, 1, numpy.array([20, 40]), numpy.array([10, 20])
    )

    assert pxl_vals[0] == 79.0 and pxl_vals[1] == 80.0


def test_set_img_pxl_values(tmp_path):
    import rsgislib.imageutils
    import numpy

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_pxl_values(
        input_img, 1, numpy.array([20, 40]), numpy.array([10, 20]), 1
    )
    pxl_vals = rsgislib.imageutils.get_img_pxl_values(
        input_img, 1, numpy.array([20, 40]), numpy.array([10, 20])
    )
    assert pxl_vals[0] == 1 and pxl_vals[1] == 1


def test_get_img_pxl_column():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    pxl_vals = rsgislib.imageutils.get_img_pxl_column(input_img, 20, 10)

    assert (
        pxl_vals[0] == 79.0
        and pxl_vals[1] == 83.0
        and pxl_vals[2] == 51.0
        and pxl_vals[3] == 48.0
        and pxl_vals[4] == 42.0
        and pxl_vals[5] == 40.0
        and pxl_vals[6] == 37.0
        and pxl_vals[7] == 37.0
        and pxl_vals[8] == 27.0
        and pxl_vals[9] == 26.0
    )


def test_assign_random_pxls(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.assign_random_pxls(
        input_img,
        output_img,
        n_pts=100,
        img_band=1,
        gdalformat="KEA",
        edge_pxl=0,
        use_no_data=True,
        rnd_seed=42,
    )

    assert os.path.exists(output_img)


def test_generate_random_pxl_vals_img(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.generate_random_pxl_vals_img(
        input_img, output_img, gdalformat="KEA", low_val=100, up_val=200
    )

    assert os.path.exists(output_img)


def test_gen_sampling_grid(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.gen_sampling_grid(
        input_img, output_img, "KEA", 10, 10, 100, True
    )

    assert os.path.exists(output_img)


def test_whiten_image(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    valid_msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.whiten_image(input_img, valid_msk_img, 1, output_img, "KEA")

    assert os.path.exists(output_img)


def test_spectral_smoothing(tmp_path):
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    valid_msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    output_img = os.path.join(tmp_path, "output_img.kea")
    rsgislib.imageutils.spectral_smoothing(input_img, valid_msk_img, 1, output_img)

    assert os.path.exists(output_img)


def test_set_img_band_metadata(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_band_metadata(input_img, 1, "Hello", "World")


def test_get_img_band_metadata():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    rsgislib.imageutils.get_img_band_metadata(input_img, 1, "LAYER_TYPE")


def test_get_img_band_metadata_fields():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    rsgislib.imageutils.get_img_band_metadata_fields(input_img, 1)


def test_get_img_band_metadata_fields_dict():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    rsgislib.imageutils.get_img_band_metadata_fields_dict(input_img, 1)


def test_set_img_metadata(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_metadata(input_img, "Hello", "World")


def test_get_img_metadata(tmp_path):
    import rsgislib.imageutils

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    input_img = os.path.join(tmp_path, "sen2_20210527_aber_subset.kea")
    copy2(input_ref_img, input_img)

    rsgislib.imageutils.set_img_metadata(input_img, "Hello", "World")
    rsgislib.imageutils.get_img_metadata(input_img, "Hello")


def test_get_img_metadata_fields():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    rsgislib.imageutils.get_img_metadata_fields(input_img)


def test_get_img_metadata_fields_dict():
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    rsgislib.imageutils.get_img_metadata_fields_dict(input_img)


def test_get_img_pxl_coords():
    import rsgislib.imageutils
    import numpy

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")

    x_coords = numpy.array([260453.28, 260430.37])
    y_coords = numpy.array([281327.15, 279581.76])

    x_pxl_coords, y_pxl_coords = rsgislib.imageutils.get_img_pxl_coords(
        input_img, x_coords, y_coords
    )

    assert (
        (x_pxl_coords[0] == 286)
        and (y_pxl_coords[0] == 151)
        and (x_pxl_coords[1] == 284)
        and (y_pxl_coords[1] == 325)
    )



#!/usr/bin/env python
"""
The elevation module contains functions for performing analysis on elevation data.
"""

# import the C++ extension into this level
from ._elevation import *


def fft_dem_fusion(highResDEMImg, lowResDEMImg, fusedDEMImg, cSize=20):
    """
    This function merges DEMs (any single band file) using a fourier
    transformation where the low frequency component is taken from the
    lower resolution DEM and the high frequency component is taken from
    the higher resolution DEM.

    This is an implementation ideas from:

    Karkee, M., Steward, B. L., & Aziz, S. A. (2008). Improving quality
    of public domain digital elevation models through data fusion.
    Biosystems Engineering, 101(3), 293–305.

    :param high_res_DEM_img: is an input image file for the high resolution DEM file.
    :param low_res_DEM_img: is an input image file for the low resolution DEM file.
    :param fused_DEM_img: is an output image file for the resulting fused DEM.
    :param c_size: is the threshold (int) defining the high and low frequencies.

    .. code:: python

        import rsgislib
        from rsgislib import elevation

        high_res_DEM_img = "N09E009_TDX_12m_EGM08.tif"
        low_res_DEM_img = "N09E009_SRTM_30m_EGM08.tif"
        fused_DEM_img = "NO9E009_SRTM_TDX_EGM08_fusion.tif"
        c_size = 10

        elevation.fft_dem_fusion(highResDEMImg, lowResDEMImg, fusedDEMImg, cSize=20)

    """
    # TODO Add checks that images are covering the same geographic area etc. Same number of pixels as well?
    import numpy
    from osgeo import gdal
    import rsgislib
    import rsgislib.imageutils

    try:
        srcDSHighRes = gdal.Open(highResDEMImg)
        srcbandHighRes = srcDSHighRes.GetRasterBand(1)
        srcbandHighResArr = srcbandHighRes.ReadAsArray()
    except Exception as e:
        print('Unable to open "' + highResDEMImg + '"')
        raise e

    try:
        srcDSLowRes = gdal.Open(lowResDEMImg)
        srcbandLowRes = srcDSLowRes.GetRasterBand(1)
        srcbandLowResArr = srcbandLowRes.ReadAsArray()
    except Exception as e:
        print('Unable to open "' + lowResDEMImg + '"')
        raise e

    fftHighRes = numpy.fft.fft2(srcbandHighResArr)
    fshiftHighRes = numpy.fft.fftshift(fftHighRes)
    magSpectHighRes = 20 * numpy.log(numpy.abs(fshiftHighRes))

    fftLowRes = numpy.fft.fft2(srcbandLowResArr)
    fshiftLowRes = numpy.fft.fftshift(fftLowRes)
    magSpectLowRes = 20 * numpy.log(numpy.abs(fshiftLowRes))

    rows, cols = srcbandLowResArr.shape
    crow, ccol = int(rows / 2), int(cols / 2)
    fshiftLowRes[
        crow - cSize : crow + cSize, ccol - cSize : ccol + cSize
    ] = fshiftHighRes[crow - cSize : crow + cSize, ccol - cSize : ccol + cSize]
    f_ishift = numpy.fft.ifftshift(fshiftLowRes)
    img_back = numpy.fft.ifft2(f_ishift)
    img_back = numpy.abs(img_back)

    rsgislib.imageutils.create_copy_img(
        highResDEMImg, fusedDEMImg, 1, -9999, "KEA", rsgislib.TYPE_32FLOAT
    )

    try:
        srcDSOutput = gdal.Open(fusedDEMImg, gdal.GA_Update)
        srcbandOutput = srcDSOutput.GetRasterBand(1)
        srcbandOutput.WriteArray(img_back)
    except Exception as e:
        print('Unable to open "' + fusedDEMImg + '"')
        raise e
    print("Completed Fusion")



import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_get_epsg_code_from_wkt_utm():
    import rsgislib.tools.projection

    wkt_str = """PROJCS["WGS 84 / UTM zone 30N",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",-3],PARAMETER["scale_factor",0.9996],PARAMETER["false_easting",500000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","32630"]]"""

    assert 32630 == rsgislib.tools.projection.get_epsg_code_from_wkt(wkt_str)


def test_get_wkt_from_epsg_code_32630():
    import rsgislib.tools.projection

    wkt_str = rsgislib.tools.projection.get_wkt_from_epsg_code(32630)
    assert (wkt_str is not None) and ("UTM zone 30N" in wkt_str)


def test_get_wkt_from_epsg_code_27700():
    import rsgislib.tools.projection

    wkt_str = rsgislib.tools.projection.get_wkt_from_epsg_code(27700)
    assert (wkt_str is not None) and ("British" in wkt_str)


def test_get_osr_prj_obj_32630():
    import rsgislib.tools.projection

    spat_ref = rsgislib.tools.projection.get_osr_prj_obj(32630)
    assert spat_ref is not None


def test_get_osr_prj_obj_27700():
    import rsgislib.tools.projection

    spat_ref = rsgislib.tools.projection.get_osr_prj_obj(27700)
    assert spat_ref is not None


def test_degrees_to_metres():
    import rsgislib.tools.projection

    x_size, y_size = rsgislib.tools.projection.degrees_to_metres(50, 0.1, 0.05)
    assert (abs(x_size - 7169.575) < 0.1) and (abs(y_size - 5561.453) < 0.1)


def test_metres_to_degrees():
    import rsgislib.tools.projection

    lon_size, lat_size = rsgislib.tools.projection.metres_to_degrees(-25, 250, 500)
    assert (abs(lon_size - 0.0025) < 0.001) and (abs(lat_size - 0.0045) < 0.001)


def test_get_deg_coord_as_str_default():
    import rsgislib.tools.projection

    coords_str = rsgislib.tools.projection.get_deg_coord_as_str(-25, 60)
    assert coords_str == "s0250e0600"


def test_get_deg_coord_as_str_chars3():
    import rsgislib.tools.projection

    coords_str = rsgislib.tools.projection.get_deg_coord_as_str(-25, 60, n_chars=3)
    assert coords_str == "s250e600"

def test_great_circle_distance_1():
    import rsgislib.tools.projection

    pt1 = [52, 70]
    pt2 = [23, 40]
    dist_val = rsgislib.tools.projection.great_circle_distance(pt1, pt2)

    assert (dist_val - 3737404.124135579) < 1

def test_great_circle_distance_2():
    import rsgislib.tools.projection

    pt1 = [23, 40]
    pt2 = [52, 70]
    dist_val = rsgislib.tools.projection.great_circle_distance(pt1, pt2)

    assert (dist_val - 3737404.124135579) < 1





#!/usr/bin/env python
"""
These utilities allow for a more 'intelligent tiling process to be carried out:

.. code:: python

    import rsgislib
    from rsgislib.imageutils import tilingutils
    
    inputImage = 'LS5TM_20110428_sref_submask_osgb.kea'
    tileShp = 'LS5TM_20110428_sref_submask_osgb_tiles.shp'
    outTilesMaskBase = 'tilesmeta/LS5TM_20110428_sref_submask_osgb_tile'
    outTilesImgBase = 'tiles/LS5TM_20110428_sref_submask_osgb_tile'
    width = 2500
    height = 2500
    validDataThreshold = 0.3
    
    tilingutils.create_min_data_tiles(inputImage, tileShp, width, height, validDataThreshold)
    tilingutils.createTileMaskImages(inputImage, tileShp, outTilesMaskBase)
    tilingutils.create_tiles_from_masks(inputImage, outTilesMaskBase, outTilesImgBase, rsgislib.TYPE_16UINT, 'KEA')

"""
############################################################################
#  tilingutils.py
#
#  Copyright 2015 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 02/04/2015
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import sys
import math
import glob
import os.path
import shutil

import rsgislib
from rsgislib import segmentation
from rsgislib import rastergis
from rsgislib import imageutils
from rsgislib import vectorutils
import rsgislib.tools.filetools

haveGDALPy = True
try:
    from osgeo import gdal, ogr
except ImportError as gdalErr:
    haveGDALPy = False

haveRIOS = True
try:
    from rios import rat
except ImportError as riosErr:
    haveRIOS = False

haveNumpy = True
try:
    import numpy
except ImportError as numpyErr:
    haveNumpy = False


def create_min_data_tiles(
    inputImage,
    outshp,
    outclumpsFile,
    width,
    height,
    validDataThreshold,
    maskIntersect=None,
    offset=False,
    force=True,
    tmpdir="tilestemp",
    inImgNoDataVal=0.0,
):
    """
    A function to create a tiling for an input image where each tile has a minimum amount of valid data.

    :param inputImage: is a string for the image to be tiled
    :param outshp: is a string for the output shapefile the tiling will be written to (if None a shapefile won't be outputted).
    :param outclumpsFile: is a string for the output image file containing the tiling
    :param width: is an int for the width of the tiles
    :param height: is an int for the height of the tiles
    :param validDataThreshold: is a float (0-1) with the proportion of valid data needed within a tile.
    :param force: is a boolean (default True) to delete the output shapefile if it already exists.
    :param tmpdir: is a string with a temporary directory for temp outputs to be stored (they will be deleted).
                  if tmpdir doesn't exist it will be created and then deleted during the processing.
    :param inImgNoDataVal: is a float for providing the input image no data value (Default: 0.0)
    """
    tmpPresent = True
    if not os.path.exists(tmpdir):
        print("WARNING: tmpdir directory does not exist so creating it...")
        os.makedirs(tmpdir)
        tmpPresent = False

    inImgBaseName = os.path.splitext(os.path.basename(inputImage))[0]

    tileClumpsImage = os.path.join(tmpdir, inImgBaseName + "_tilesimg.kea")

    segmentation.generateRegularGrid(
        inputImage, tileClumpsImage, "KEA", width, height, offset
    )
    rastergis.pop_rat_img_stats(tileClumpsImage, True, True)

    if not maskIntersect is None:
        bs = []
        bs.append(rastergis.BandAttStats(band=1, maxField="Mask"))
        rastergis.populate_rat_with_stats(maskIntersect, tileClumpsImage, bs)
        ratDS = gdal.Open(tileClumpsImage, gdal.GA_Update)
        MaskField = rat.readColumn(ratDS, "Mask")
        Selected = numpy.zeros_like(MaskField, dtype=int)
        Selected[MaskField == inImgNoDataVal] = 1
        rat.writeColumn(ratDS, "Selected", Selected)
        ratDS = None

        tileClumpsImageDropClumps = os.path.join(
            tmpdir, inImgBaseName + "_tilesimgdropped.kea"
        )

        segmentation.dropSelectedClumps(
            tileClumpsImage, tileClumpsImageDropClumps, "KEA", "Selected"
        )
        rsgislib.tools.filetools.delete_file_silent(tileClumpsImage)
        tileClumpsImage = tileClumpsImageDropClumps

    rastergis.populateRATWithPropValidPxls(
        inputImage, tileClumpsImage, "ValidPxls", inImgNoDataVal
    )

    ratDS = gdal.Open(tileClumpsImage, gdal.GA_Update)
    ValidPxls = rat.readColumn(ratDS, "ValidPxls")
    Selected = numpy.zeros_like(ValidPxls, dtype=int)
    NoDataClumps = numpy.zeros_like(ValidPxls, dtype=int)
    Selected[ValidPxls < validDataThreshold] = 1
    NoDataClumps[ValidPxls == 0] = 1
    Selected[ValidPxls == 0] = 0
    rat.writeColumn(ratDS, "Selected", Selected)
    rat.writeColumn(ratDS, "NoDataClumps", NoDataClumps)
    ratDS = None

    segmentation.mergeSegments2Neighbours(
        tileClumpsImage, inputImage, outclumpsFile, "KEA", "Selected", "NoDataClumps"
    )

    if not outshp is None:
        tilesDS = gdal.Open(outclumpsFile, gdal.GA_ReadOnly)
        tilesDSBand = tilesDS.GetRasterBand(1)

        dst_layername = os.path.splitext(os.path.basename(outshp))[0]
        # print(dst_layername)
        drv = ogr.GetDriverByName("ESRI Shapefile")

        if force and os.path.exists(outshp):
            drv.DeleteDataSource(outshp)

        dst_ds = drv.CreateDataSource(outshp)
        dst_layer = dst_ds.CreateLayer(dst_layername, srs=None)

        gdal.Polygonize(tilesDSBand, tilesDSBand, dst_layer, -1, [], callback=None)

        tilesDS = None
        dst_ds = None

    if not tmpPresent:
        shutil.rmtree(tmpdir, ignore_errors=True)
    else:
        rsgislib.tools.filetools.delete_file_silent(tileClumpsImage)


def create_tile_mask_images_from_shp(
    inputImage,
    tileShp,
    tilesNameBase,
    tilesMaskDIR,
    tmpdir="tilestemp",
    imgFormat="KEA",
):
    """
    A function to create individual image masks from the tiles shapefile which can be
    individually used to mask (using rsgislib mask function) each tile from the inputimage.

    :param inputImage: is the input image being tiled.
    :param tileShp: is a shapefile containing the shapefile tiles.
    :param tilesNameBase: is the base file name for the tile masks
    :param tilesMaskDIR: is the directory where the output images will be outputted
    :param tmpdir: is a string with a temporary directory for temp outputs to be stored (they will be deleted)
                   If tmpdir doesn't exist it will be created and then deleted during the processing.
    """

    tmpPresent = True
    if not os.path.exists(tmpdir):
        print("WARNING: tmpdir directory does not exist so creating it...")
        os.makedirs(tmpdir)
        tmpPresent = False

    shpTilesBase = os.path.join(tmpdir, tilesNameBase + "_tileshp")

    vectorutils.splitFeatures(tileShp, shpTilesBase, True)

    drv = ogr.GetDriverByName("ESRI Shapefile")
    shpFiles = glob.glob(shpTilesBase + "*.shp")
    idx = 1
    for shpFile in shpFiles:
        imgTileFile = os.path.join(tilesMaskDIR, tilesNameBase + str(idx) + ".kea")
        shpFileLyr = os.path.splitext(os.path.basename(shpFile))[0]
        rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
            shpFile,
            shpFileLyr,
            inputImage,
            imgTileFile,
            gdalformat=imgFormat,
            burn_val=1,
            datatype=rsgislib.TYPE_8UINT,
            att_column=None,
            use_vec_extent=False,
            thematic=True,
            no_data_val=0,
        )
        drv.DeleteDataSource(shpFile)
        idx = idx + 1

    if not tmpPresent:
        shutil.rmtree(tmpdir, ignore_errors=True)


def create_tile_mask_images_from_clumps(
    clumpsImage, tilesNameBase, tilesMaskDIR, gdalformat="KEA"
):
    """
    A function to create individual image masks from the tiles shapefile which can be
    individually used to mask (using rsgislib mask function) each tile from the inputimage.

    :param clumpsImage: is an image file with RAT where each clump represented a tile region.
    :param tilesNameBase: is the base file name for the tile masks
    :param tilesMaskDIR: is the directory where the output images will be outputted
    :param gdalformat: is the output image file format of the tile masks
    """
    import rsgislib.tools.filetools

    outBaseImg = os.path.join(tilesMaskDIR, tilesNameBase)
    outImgExt = rsgislib.imageutils.get_file_img_extension(gdalformat)[1:]
    rastergis.exportClumps2Images(
        clumpsImage, outBaseImg, True, outImgExt, gdalformat, 1
    )


def create_tiles_from_masks(
    inputImage, tilesBase, tilesMetaDIR, tilesImgDIR, datatype, gdalformat
):
    """
    A function to apply the image tile masks defined in createTileMaskImages to the input image to extract the individual tiles.

    :param inputImage: is the input image being tiled.
    :param tileMasksBase: is the base path for the tile masks. glob will be used to find them with \*.kea added to the end.
    :param outTilesBase: is the base file name for the tiles.

    """
    maskFiles = glob.glob(os.path.join(tilesMetaDIR, tilesBase + "*.kea"))

    idx = 1
    for maskFile in maskFiles:
        tileImage = os.path.join(tilesImgDIR, os.path.basename(maskFile))
        imageutils.mask_img(inputImage, maskFile, tileImage, gdalformat, datatype, 0, 0)
        imageutils.pop_img_stats(tileImage, True, 0.0, True)
